// -------------------------------------------------------------------------------------
// 3DTouch2
// Alternative Firmware fuer 3DTouch Autolevelling Sensor
// Harald Sattler, 03/2018
// -------------------------------------------------------------------------------------
//
// Der 3DTouch Sensor ist dafuer gedacht, die Bodenplatte eines 
// 3D-Druckers auszumessen um eine eventuelle Schieflage der 
// Platte auszugleichen. Der Hersteller bietet Anleitungen fuer 
// die Adaption des Sensors an verschiedene 3D-Drucker sowie 
// Patch-Anleitungen um die 3D-Drucker-Software fuer den Gebrauch
// des Sensors zu aktualisieren.
// GRBL bietet keine Unterstuetzung fuer den 3DTouch, um nichtleitende 
// Oberflaechen auch mit einer durch GRBL befeuerten CNC Fraese abtasten zu
// koennen, habe ich die Schaltung analysiert und diese alternative 
// Firmware erstellt.
// Im Wesentlichen reduziert sich der Einsatz des Sensors fuer GRBL
// auf das Abtasten ueber den im Sensor eingebauten Hall-Effekt-Sensor
// und die Erzeugung des "Probe"-Impulses bei Beruehrung der Oberflaeche.
// Die Ansteuerung des Sensors ueber das Servo-Signal entfaellt komplett,
// der Taststift wird abhaengig vom Probe-Signal zurueckgezogen oder 
// ausgefahren. Das Ganze erfolgt zeitgesteuert.
// Der initiale Selbsttest wurde nachempfunden, bei Erkennung eines
// Fehlers (der Taststift bewegt sich nicht wie erwartet), blinkt die 
// rote LED am Sensor und es muss ein weiterer Selbsttest-Zyklus 
// eingeleitet werden, indem der Taststift manuell betätigt wird.
// (ggf. nach Behebung des Problems - Taststift bewegt sich nicht frei)
//
// Die alternative Firmware wird von mir fuer den Eigengebrauch des
// Anwenders zur Verfuegung gestellt, die Hardware muss weiterhin vom
// Hersteller bezogen werden, die urspruengliche Funktion des Sensors
// geht bei Aufspielen der Alternativ-Firmware unwiderruflich verloren!!!
// (der original Flash-Inhalt kann nicht ausgelesen/gesichert werden)
//
// Die alternative Firmware aufzuspielen ist kein Kinderspiel, bitte nur 
// in Angriff nehmen, wenn "SMD", "0402" und "Feinleitertechnik" keine 
// boehmischen Doerfer darstellen.
// 
// Verwendung auf eigene Gefahr.
//
// ACHTUNG!
// -------------------------------------------------------------------------------------
// Beim Programmieren muss der Taststift herausgezogen sein (ggf. festhalten)!
// Der Hall-Sensor erzeugt (im 3DTouch) keine Impulse sondern statische Signale:
//
// Stift herausgezogen -> Hall-Sensor-Ausgang = HIGH
// Stift eingezogen    -> Hall-Sensor-Ausgang = LOW
//
// Der Ausgang des Hall-Sensors ist OpenCollector, somit keine Probleme mit dem 
// Programmiersignal am gleichen Pin (Eingang des AtTiny), solange der Pin draussen ist
// -------------------------------------------------------------------------------------
//
// 3DTouch V2

// #define LED 5				// PB5 - Pin 1	// Originalbeschaltung
// #define SERVO 0				// PB0 - Pin 5	// Originalbeschaltung
#define LED 0					// PB0 - Pin 5	// neue Beschaltung
#define SENSOR 1				// PB1 - Pin 6
#define ZPROBE 2				// PB2 - Pin 7
#define PINUP 3					// PB3 - Pin 2	// PD3-Pin 1 beim Nano
#define PINDOWN 4				// PB4 - Pin 3	// PD4-Pin 2 beim Nano
#define DELAYNACHPIN 100		// Zeit[ms], in der das Touch-Event erfolgt sein muss

// um Irritationen wegen AN = LOW und AUS = HIGH zu vermeiden...
#define LED_AN 	(digitalWrite(LED, LOW))
#define LED_AUS (digitalWrite(LED, HIGH))
// Hinweis: "LED_AN" ist 80 ns langsamer als "PORTB = PINB &= ~(0x01 << 0))" aber
//			6 Byte kuerzer, daher werden die Makros verwendet

// globale Variablen
volatile bool sensorTripped;	// Sensor wurde getriggert, wird im IRQ gesetzt
bool flashLED = false;			// (bisher) kein Fehler aufgetreten
bool selbstTest = true;			// Selbsttest (noch) nicht erfolgreich abgeschlossen
bool pinIsDown;					// Status des Pin
int i;

uint32_t lastLoopTimer;
uint32_t tripTimer;
uint32_t lastTriptimer = 0;
// bool firstTrip = false;				// fuer die Zeitmessung

/* void blinkBits(uint8_t bitData)
// gehoert zu "Fuses auslesen"
{
	for( byte iii = 0; iii<8; iii++)
	{
		if((bitData & (0x01 << iii)))
		{								// LED lange an
			LED_AN;
			delay(50);
			LED_AUS;
			delay(500);
		}
		else
		{								// LED kurz an
			LED_AN;
			delay(950);
			LED_AUS;
			delay(500);
		}
	}
}
 *///-------------------------------------------------------------
void setup()
{
	// Pins konfigurieren und inaktiv schalten
	pinMode(LED, OUTPUT);
	LED_AUS;							// aktiv LOW Ausgang -> LED aus
	pinMode(PINUP, OUTPUT);
	digitalWrite(PINUP, HIGH);			// aktiv LOW Ausgang deaktivieren
	pinMode(PINDOWN, OUTPUT);
	digitalWrite(PINDOWN, HIGH);		// aktiv LOW Ausgang deaktivieren
	pinMode(ZPROBE, OUTPUT);
	digitalWrite(ZPROBE, HIGH);			// aktiv LOW Ausgang deaktivieren
	pinMode(SENSOR, INPUT_PULLUP);
	
//{ --------------- Fuses auslesen --------------- 
// Dieser Codeteil liest die Fuses und Lockbits aus und morst sie aufder LED raus
// Dazu gehoert die Subroutine blinkBits(), die ebenfalls einkommentiert werden muss
/*
// LOW  Fuses, High Fuses, 	Lockbits auslesen und anzeigen per Blinken
#define GET_LOW_FUSE_BITS           (0x0000)
#define GET_LOCK_BITS               (0x0001)
#define GET_EXTENDED_FUSE_BITS      (0x0002)
#define GET_HIGH_FUSE_BITS          (0x0003)
#define SELFPRGEN 4
#define __BOOT_LOCK_BITS_SET      (_BV(SELFPRGEN) | _BV(RFLB))

#define boot_lock_fuse_bits_get(address)                   \
(__extension__({                                           \
    uint8_t __result;                                      \
    __asm__ __volatile__                                   \
    (                                                      \
        "ldi r30, %3\n\t"                                  \
        "ldi r31, 0\n\t"                                   \
        "sts %1, %2\n\t"                                   \
        "lpm %0, Z\n\t"                                    \
        : "=r" (__result)                                  \
        : "i" (_SFR_MEM_ADDR(SPMCSR)),                  \
          "r" ((uint8_t)__BOOT_LOCK_BITS_SET),             \
          "M" (address)                                    \
        : "r0", "r30", "r31"                               \
    );                                                     \
    __result;                                              \
}))

uint8_t data1,data2,data3,data4;

cli();
data1 = boot_lock_fuse_bits_get(GET_LOW_FUSE_BITS);
data2 = boot_lock_fuse_bits_get(GET_HIGH_FUSE_BITS);
// data3 = boot_lock_fuse_bits_get(GET_EXTENDED_FUSE_BITS);
data4 = boot_lock_fuse_bits_get(GET_LOCK_BITS);
sei();	

delay(2000);
for(i=0; i<40;i++)
{
	LED_AN;
	delay(20);
	LED_AUS;
	delay(20);
}
delay(2000);

blinkBits(0x55);
delay(2000);

blinkBits(data1);
delay(2000);
blinkBits(data2);

delay(2000);
blinkBits(data4);

while(1){};		// hier anhalten
*///} --------------- Fuses auslesen --------------- 

	flashLED = false;			// (bisher) kein Fehler aufgetreten
	selbstTest = true;			// Selbsttest (noch) nicht erfolgreich abgeschlossen
	
	// Interrupt fuer Sensor zuweisen
	attachInterrupt(digitalPinToInterrupt(SENSOR), touchINT, FALLING);	//CHANGE
	
	sensorTripped = false;			// Flag ruecksetzen

	//{ einer der FETs hat bei den HV-Prog-Versuchen das Zeitliche gesegnet, ich
	// musste also eine Moeglichkeit schaffen, die H-Bruecke aus 4 FETs zu testen.
	// debug - H-Bruecke messen
	// for(i=0; i<200; i++)
	// {
		// pinDown();						// Pin absenken
		// delay(500);						// keine Hektik...

		// pinUp();						// Pin hochziehen
		// delay(500);						// keine Hektik...

		// digitalWrite(LED, !digitalRead(LED));

	// }	//}

	// Selbsttest starten, der Pin muss zweimal hoch und runter gehen,
	// jedesmal muss ein Sensor-Event ausgeloest werden
	while(selbstTest == true)
	{
		LED_AUS;			// LED aus
		
		// Selbsttest durchfuehren (2mal Pin runter und hoch, 
		//   dabei Check ob die INTs vom Sensor kommen)
		for(i=0; i<2; i++)
		{
			pinDown();						// Pin absenken
			if(sensorTripped == true)		// der Sensor-Interrupt muss gekommen sein
			{
				sensorTripped = false;		// ok -> Flag ruecksetzen
				flashLED = false;			// kein Fehler
			}
			else							// kein IRQ -> FEHLER
			{
				flashLED = true;			// LED blinkt hektisch -> Fehler
				// selbstTest = true;			// nochmal testen
			}
			delay(500);						// keine Hektik...
			
			pinUp();						// Pin hochziehen
			if(sensorTripped == true)		// der Sensor-Interrupt muss gekommen sein
			{
				sensorTripped = false;		// ok -> Flag ruecksetzen
				flashLED = false;			// kein Fehler
			}
			else							// kein IRQ -> FEHLER
			{
				flashLED = true;			// LED blinkt hektisch -> Fehler
				// selbstTest = true;		// nochmal testen
			}
			delay(500);						// keine Hektik...
		}

		if(flashLED == true)				// es ist ein Fehler aufgetreten
		{	
			// nach Fehler solange blinken, bis der Pin manuell betätigt wurde
			while(sensorTripped == false)	
			{
				PINB = bit (LED); // toggle B2
				delay(100);
			}
			
			flashLED = false;				// Blinken stoppen
			sensorTripped = false;			// Flag ruecksetzen
			
			// ab hier wieder an den Anfang von Selbsttest
			selbstTest = true;				// nochmal testen
		}
		else								// beim Test kein Fehler aufgetreten -> ok
		{
			flashLED = false;				// Blinken stoppen
			selbstTest = false;				// alles ok, weiter in loop()
			pinDown();						// Pin in Taststellung bringen
			pinIsDown = true;				// der Pin ist jetzt draussen
			
			// LED zweimal kurz aufblitzen lassen
			for(i=0; i<3;i++)
			{
				LED_AN;
				delay(50);				
				LED_AUS;
				delay(250);				
			}
			sensorTripped = false;			// Flag ruecksetzen
		}
	}
	digitalWrite(ZPROBE, HIGH);				// Meldung an GRBL deaktivieren
}

//-------------------------------------------------------------
void loop()
{
	
	// Schleifenzeit bestimmen 
	// PINB = bit (2); // toggle B2

 	if(sensorTripped == true)
	{
		if(pinIsDown == true)			// der Pin sollte jetzt draussen sein (1.Trip)
		{
			// Das Hochziehen des Pin setzt erneut das InterruptFlag, das darf aber _nicht_
			// an GRBL gemeldet werden. Daher: Interruptbehandlung sperren
			detachInterrupt(digitalPinToInterrupt(SENSOR));
			
			pinUp();						// Pin schnell hochziehen		+ 100 ms
			pinIsDown = false;				// der Pin sollte jetzt oben sein
			// Die Ints sind gesperrt, aber das Flag wird gespeichert und wuerde nach 
			// Freigeben der Interruptbehandlung sofort bearbeitet werden. Daher:
			// Pin Change Interrupt Flag loeschen (S. Datenblatt: auf 1 setzen)
			GIFR |= (0x01 << PCIF);			// PCIF loeschen (indem ein 1 geschrieben wird)

			tripTimer = millis();			// Zeit merken
		}
		else
		{
			// Schleifenzeit bestimmen 
			// PINB = bit (2); // toggle B2
			lastTriptimer = millis();			// Zeit merken
			
			if(lastTriptimer > (tripTimer + 100))	// delay fuer LED, ggf. anpassen
			{
				LED_AUS;						// LED wieder aus
				digitalWrite(ZPROBE, HIGH);		// Meldung an GRBL deaktivieren
			}
						
			// if(lastTriptimer > (tripTimer + 200))	// delay fuer Meldung an GRBL
			// {
				// digitalWrite(ZPROBE, HIGH);		// Meldung an GRBL deaktivieren
			// }
			
			if(lastTriptimer > (tripTimer + 500))	// delay fuer Pin, ggf. anpassen
			{
				pinDown();						// Pin wieder in Taststellung	+ 100 ms
				// Das Ausfahren des Pin setzt erneut das InterruptFlag, das darf aber _nicht_
				// an GRBL gemeldet werden. Die Ints sind gesperrt, aber das Flag wird 
				// gespeichert und wuerde nach Freigeben der Interruptbehandlung sofort 
				// bearbeitet werden. Daher:
				// Pin Change Interrupt Flag loeschen (S. Datenblatt: auf 1 setzen)
				GIFR |= (0x01 << PCIF);			// PCIF loeschen (indem ein 1 geschrieben wird)
				
				pinIsDown = true;				// der Pin sollte jetzt draussen sein
				
				// erst jetzt duerfen wieder Interrupts erfolgen
				attachInterrupt(digitalPinToInterrupt(SENSOR), touchINT, FALLING);	//CHANGE, LOW
				
				sensorTripped = false;
			}
		}
	}

/*	// Fuer den nachfolgenden Code ist leider kein Platz im AtTiny13, dafuer muesste
	// ein AtTiny23 oder so herhalten.
	// Durch oben festhalten des Pin fuer >1s wird der Sensor abgeschaltet (Sleepmode)
	if((pinIsDown == true) && (digitalRead(SENSOR) == LOW))
	{
		for(i=0; i<1; i++)			// LED zweimal kurz aufblitzen lassen
		{
			LED_AN;
			delay(50);				
			LED_AUS;
			delay(250);				
		}
		
		for (j=0; j <= 255; j += 5) // fade out
		{
			analogWrite(LED, j);	// sets the value (range from 0 to 255):
			delay(30);	// wait for 30 milliseconds to see the dimming effect		
		}
		sleeping = true;
	}
	else
	{
		pinIsDown = false;
	}
	// else		// wir sind im sleep mode
	// {
		// if(firstTrip == true)	// nichts tun, auf den naechsten Trip warten
		// {
			// for (j=255; j >= 0; j -= 5) // fade in
			// {
				// analogWrite(LED, j);	// sets the value (range from 0 to 255):
				// delay(30);	// wait for 30 milliseconds to see the dimming effect		
			// }
			// for(i=0; i<1;i++)	// LED zweimal kurz aufblitzen lassen
			// {
				// LED_AN;
				// delay(50);				
				// LED_AUS;
				// delay(250);				
			// }
		// }
	// }
*/	
}

//-------------------------------------------------------------
//{ Interruptroutine fuer Sensor
void touchINT()
{
	// jetzt so schnell wie moeglich die Meldung zu GRBL senden, ZPROBE muss LOW werden
	// gleichzeitig setzen wir die LED auf AN (beide Signale sind LOW aktiv!!)
	// Beim Tiny erfolgt das _gleichzeitig_ (nur Port B vorhanden)
	// Bit 0 -> LED,  Bit 2 -> Ausgang fuer GRBL
	// PORTB = PINB &= ~((0x01 << 2) | (0x01 << 0));	// Bits loeschen (wegen aktiv LOW)
	// Hinweis: "LED_AN" ist 80 ns langsamer als "PORTB = PINB &= ~(0x01 << 0))" aber
	//			6 Byte kuerzer, daher werden die Makros verwendet
	digitalWrite(ZPROBE, LOW);			// aktiv LOW Ausgang deaktivieren
	LED_AN;
	
	sensorTripped = true; 				// Merker, dass die Oberflaeche beruehrt wurde
}

//-------------------------------------------------------------
// Subroutinen
void pinDown()
{
	digitalWrite(PINDOWN, LOW);			// aktiv LOW Ausgang
	delay(100);							// Wert aus Original Firmware
	digitalWrite(PINDOWN, HIGH);		// aktiv LOW Ausgang
}

void pinUp()
{
	digitalWrite(PINUP, LOW);			// aktiv LOW Ausgang
	delay(100);							// Wert aus Original Firmware
	digitalWrite(PINUP, HIGH);			// aktiv LOW Ausgang
}
//}-------------------------------------------------------------
